﻿#include "stdafx.h"
#include "Connector.h"
#include <MSWSock.h>
#include "Peer.h"
#include "IocpThreadManager.h"
#include "SessionManager.h"

HANDLE CConnector::m_hIOCP = INVALID_HANDLE_VALUE;
int CConnector::m_nRefCount = 0;

/**
@brief 생성자
@param peer IO를 담당할 피어를 생성자에서 전달 받는다.
@param bActive 시작과 동시에 바로 스타트 유효하게 동작 시킬껀지 여부.
*/
CConnector::CConnector(CPeer* pcPeer, bool bActive) : m_lpfnConnectEx(NULL), m_bTryConnect(false), m_bActive(bActive)
{
	m_nRefCount++;
	GetSessionManager()->New(1);
	m_pcSession.reset(GetSessionManager()->Pop());

	SetPeer(pcPeer);
}

/**
@brief 소멸자. 각종 릴리즈 작업을 한다.
*/
CConnector::~CConnector()
{
	m_pcSession->Release();
	GetSessionManager()->Push(m_pcSession.release()); //소유권 포기하고 큐에 넣음

	if(--m_nRefCount == 0)
	{
		FreeThread();
	}
}

/**
@brief 쓰레드 할당 메소드
@param nThreadCnt 이 파라미터 만큼 스레드를 생성한다.
*/
void CConnector::AllocThread(int nThreadCnt)
{
	if(m_hIOCP == INVALID_HANDLE_VALUE)
	{
		m_hIOCP = CreateIOCPHandle();
	}
	GetSmartSingleton()->Register(m_hIOCP, nThreadCnt);
}

/**
@brief Listener가 사용하는 모든 쓰레드를 해제하는 메소드.
*/
void CConnector::FreeThread()
{
	GetSmartSingleton()->Unregister(m_hIOCP);
}

/**
@brief 일정 시간 이상 패킷이 오지 않은 CSession를 Disconnect 해버린다.
*/
void CConnector::AliveCheck()
{
	DWORD curtick = GetTickCount();
	if(m_pcSession->isConnected() && !m_pcSession->CheckTick(curtick))
	{
		LOG_INFO(_T("%s CheckTick() False [%d, %d]"), __TFUNCTION__, m_pcSession->GetIdx(), m_pcSession->GetAccKey());
		Disconnect();
	}
}

/**
@brief 접속시에는 CPeer의 Act를, 접속중이 아닐때는 접속을 시도하는 메소드
*/
void CConnector::Act()
{
	if(!m_bActive)	
		return;

	if(isConnected())
	{
		m_pcPeer->Act();
	}
	else
	{
		Connect();
	}
}

/**
@brief Disconnect을 시도하는 메소드.
*/
void CConnector::Disconnect()
{
	Disconnect(m_pcSession.get());
}

/**
@brief Disconnect을 시도하는 메소드
@param pcSession 접속을 끊을 pcSocketConext
*/
void CConnector::Disconnect(CSession* pcSession)
{
	if(pcSession->isConnected())
	{
		LOG_INFO(_T("%s [%d, %d]"), __TFUNCTION__, pcSession->GetIdx(), pcSession->GetAccKey());
		pcSession->CloseSocket(); //여기서 클로즈 소켓하면 GQCS에서 리턴됨.
	}
	else
	{
		LOG_INFO(_T("%s Failed [%d, %d]"), __TFUNCTION__, pcSession->GetIdx(), pcSession->GetAccKey());
	}
}

/**
@brief 접속이 끊겼을 시 불려진다.
@param pcSession 접속이 끊긴 CSession
*/
void CConnector::OnDisconnect(CSession* pcSession)
{
	pcSession->Enter();
	if(pcSession->isConnected())
	{
		LOG_INFO(_T("%s [Close Reason : %d] [%d, %d] [%s:%d]"), __TFUNCTION__, WSAGetLastError(), m_pcSession->GetIdx(), m_pcSession->GetAccKey(), m_pcSession->GetIP().c_str(), m_pcSession->GetPort());
		pcSession->OnClose();
		pcSession->CloseSocket();
		pcSession->Initialize();
		_Set();
	}
	else if(m_bTryConnect)
	{
		LOG_INFO(_T("%s [ConnectFailed : %d] [%d, %d] [%s:%d]"), __TFUNCTION__, WSAGetLastError(), m_pcSession->GetIdx(), m_pcSession->GetAccKey(), m_pcSession->GetIP().c_str(), m_pcSession->GetPort());
		pcSession->CloseSocket();
		pcSession->Initialize();
		_Set();
		InterlockedExchange((LONG *)(&m_bTryConnect), (LONG)false);
	}
	else
	{
		LOG_INFO(_T("%s [%d, %d]"), __TFUNCTION__, pcSession->GetIdx(), pcSession->GetAccKey());
	}
	pcSession->Leave();
}

/**
@brief소켓을 초기화하고 옵션을 설정하는 메소드
@return 성공 여부
*/
bool CConnector::_Set()
{
	if(m_pcSession->isConnected())
		return false;

	m_pcSession->Enter();
	SOCKET hSocket = WSASocket(AF_INET, SOCK_STREAM, IPPROTO_TCP, NULL, 0, WSA_FLAG_OVERLAPPED);
	if(INVALID_SOCKET == hSocket)
	{
		LOG_ERROR(_T("%s WSASocket() Fail. [%d]"), __TFUNCTION__, WSAGetLastError());
		m_pcSession->Leave();
		return false;
	}

	DWORD dwBytes;
	GUID GuidConnectEx = WSAID_CONNECTEX;
	if (SOCKET_ERROR  == ::WSAIoctl( hSocket, SIO_GET_EXTENSION_FUNCTION_POINTER, &GuidConnectEx, sizeof(GuidConnectEx), &m_lpfnConnectEx, sizeof(m_lpfnConnectEx) , &dwBytes, NULL, NULL))
	{
		LOG_ERROR(_T("%s WSAIoctl() Fail [%d]"), __TFUNCTION__, WSAGetLastError());
		CSocketInterface::CloseSocket(hSocket);
		m_pcSession->Leave();
		return false;
	}

	int zero = 0;
	if( SOCKET_ERROR == setsockopt(hSocket, SOL_SOCKET, SO_SNDBUF, (char*)&zero, sizeof(zero)) )
	{	  	   
		LOG_ERROR(_T("%s setsockopt() SO_SNDBUF error. [%d]"), __TFUNCTION__, WSAGetLastError());
		CSocketInterface::CloseSocket(hSocket);
		m_pcSession->Leave();
		return false;
	}

	if( SOCKET_ERROR == setsockopt(hSocket, SOL_SOCKET, SO_RCVBUF, (char*)&zero, sizeof(zero)) )
	{	  	   
		LOG_ERROR(_T("%s setsockopt() SO_RCVBUF error. [%d]"), __TFUNCTION__, WSAGetLastError());
		CSocketInterface::CloseSocket(hSocket);
		m_pcSession->Leave();
		return false;
	}

	// Note : 소켓주소구조체를 구성해서 bind작업을 한다. 주소구조체 구성시에는 Local주소중 하나를 사용하고, Protocol family만 지정해주면 된다.
	SOCKADDR_IN	localAddr;
	::ZeroMemory(&localAddr, sizeof(SOCKADDR_IN)); 
	localAddr.sin_family = AF_INET;   
	localAddr.sin_addr.s_addr = inet_addr(GetIP().c_str());	
	if( SOCKET_ERROR == ::bind( hSocket,( struct sockaddr * )&localAddr,sizeof(localAddr) ) )
	{
		LOG_ERROR(_T("%s bind() error. [%d]"), __TFUNCTION__, WSAGetLastError());
		CSocketInterface::CloseSocket(hSocket);
		m_pcSession->Leave();
		return false;
	}

	m_pcSession->SetSocket(hSocket);
	m_pcSession->SetWorker(this);
	m_pcSession->SetPeer(m_pcPeer.get());
	m_pcSession->Leave();
	return true;
}

/**
@brief 접속을 시도하는 메소드
@return 성공 여부
*/
bool CConnector::Start()
{ 
	if(m_hIOCP == INVALID_HANDLE_VALUE)
	{
		LOG_ERROR(_T("%s Fail. m_hIOCP == INVALID_HANDLE_VALUE. You have to call AllocThread Method."), __TFUNCTION__);
		return false;
	}

	m_bActive = true;
	return Connect();
}

/**
@brief 접속 대상의 ip와 port를 설정하는 메소드.
@param szIP 접속 대상의 IP
@param nPort 접속 대상의 Port
*/
bool CConnector::Set(const char* szIP, int nPort)
{
	m_pcSession->SetSockAddr(&GetSockAddr(szIP, nPort));
	return _Set();
}

/**
@brief 접속중인 Peer들에게 패킷을 브로드 캐스트한다.
@param nIdx 찾을 CSession의 인덱스
@param nAccKey 찾을 CSession의 누적 키 인덱스
@param nLen pData의 크기
@param pData 기록할 데이터
@param nAddLen pAddData의 크기
@param pAddData 기록할 추가 데이터
@return 성공 여부
@desc pData가 버퍼의 앞쪽, pAddData가 pData의 뒤쪽에 기록된다.
*/
bool CConnector::Send(const int nLen, const void* pData, const int nAddLen /* = 0 */, const void* pAddData /* = 0 */)
{
	if(m_pcSession->isConnected())
	{
		return m_pcPeer->Send(nLen, pData, nAddLen, pAddData);
	}
	return false;
}

/**
@brief 접속을 시도하는 메소드 (블러킹 함수)
@return 성공 여부
*/
bool CConnector::Connect()
{
	if(m_bTryConnect)
		return false;
	
	InterlockedExchange((LONG *)(&m_bTryConnect), (LONG)true);
	m_pcSession->Enter();

	HANDLE hIOCP = CreateIoCompletionPort((HANDLE)m_pcSession->GetSocket(), m_hIOCP, (ULONG_PTR)m_pcSession.get(), 0);
	if (NULL == hIOCP)
	{
		if(WSAGetLastError() == ERROR_INVALID_PARAMETER)
		{
			LOG_DEBUG(_T("%s ERROR_INVALID_PARAMETER. Already Connected"), __TFUNCTION__);
			m_pcSession->Leave();
			return true; //Already Connected.
		}

		LOG_ERROR(_T("%s CreateIoCompletionPort Fail : %d, %d, %d, %p"), __TFUNCTION__, WSAGetLastError(), m_pcSession->GetSocket(), m_hIOCP, m_pcSession.get());
		OnDisconnect(m_pcSession.get());
		m_pcSession->Leave();
		return false;
	}

	SOCKADDR_IN* pSockAddrIn = m_pcSession->GetSockAddrIn();
	CIocpContext* pAcceptContext = m_pcSession->GetAcceptIocpContext();

	BOOL bRet = m_lpfnConnectEx(m_pcSession->GetSocket(), (struct sockaddr*)pSockAddrIn, sizeof(struct sockaddr), NULL, 0, NULL, pAcceptContext->GetOverlappedPtr());
	if(!bRet && WSAGetLastError() != WSA_IO_PENDING)
	{
		LOG_ERROR(_T("%s m_lpfnConnectEx Fail: %d"), __TFUNCTION__, WSAGetLastError());
		OnDisconnect(m_pcSession.get());
		m_pcSession->Leave();
		return false;
	}

	m_pcSession->Leave();
	return true;
}


/**
@brief 접속 성공시 불려지는 메소드.
@param pcSession 접속 성공한 CSocketConext
*/
void CConnector::OnConnected(CSession* socketcontext)
{
	LOG_DEBUG(_T("%s [%d, %d]"), __TFUNCTION__, socketcontext->GetIdx(), socketcontext->GetAccKey());

	InterlockedExchange((LONG *)(&m_bTryConnect), (LONG)false);

	m_pcSession->OnConnect();

	m_pcSession->Enter();
	m_pcSession->Receive();
	m_pcSession->Leave();
}

/**
@brief Send 성공시 불려지는 메소드.
@param pcSession Send 성공한 CSocketConext
*/
void CConnector::OnSend(CSession* socketcontext)
{
	LOG_DEBUG(_T("%s [%d, %d]"), __TFUNCTION__, socketcontext->GetIdx(), socketcontext->GetAccKey());
}

/**
@brief Receive 성공시 불려지는 메소드.
@param pcSession Receive 성공한 CSocketConext
*/
void CConnector::OnReceive(CSession* socketcontext)
{
	LOG_DEBUG(_T("%s [%d, %d]"), __TFUNCTION__, socketcontext->GetIdx(), socketcontext->GetAccKey());
}

/**
@brief Io실패시 불려지는 메소드
@param pcSession Io가 실패한 CSession
*/
void CConnector::OnIoFailed(CSession* pcSession)
{
	OnDisconnect(pcSession);
}

/**
@brief 피어를 설정하는 메소드
@param pcPeer 설정할 피어
*/
void CConnector::SetPeer(CPeer* pcPeer)
{
	m_pcPeer.reset(pcPeer);	
	m_pcSession->SetPeer(m_pcPeer.get());
}

/**
@brief 커넥터를 멈추게 한다.
*/
void CConnector::Stop()
{
	if(isConnected())
		Disconnect(m_pcSession.get());

	InterlockedExchange((LONG*)&m_bActive, (LONG)false);
}