﻿#include "stdafx.h"
#include "Session.h"
#include "Peer.h"
#include "Tick.h"
#include <MSWSock.h> 
#include "PacketQueue.h"
#include "IocpInterface.h"
#include "MiniDump.h"

/**
@brief 생성자. 각종 초기화 작업을 한다
@param nIdx 소켓 식별값. 상수로써 변하지 않는다.
@param pcWorker IOCP기능을 해주는 Worker.
*/
CSession::CSession(const int nIdx, CIocpInterface* pcWorker) : m_bConnected(false), m_nIdx(nIdx)
{
	SetWorker(pcWorker);
	memset(&m_SockAddrIn, 0, sizeof(m_SockAddrIn));
	m_pcTick.reset(new CTick);
}

/**
@brief 소멸자. Release메소드를 호출 해줌으로써 릴리즈 동작을 한다.
*/
CSession::~CSession()
{
	Release();
}

/**
@brief 접속 종료를 시도한다.
*/
void CSession::Disconnect()
{
	if(!isConnected())
	{
		LOG_INFO(_T("%s !isConnected() [%d, %d]"), __TFUNCTION__, GetIdx(), GetAccKey());
		return;
	}

	if(m_pcWorker)
	{
		m_pcWorker->Disconnect(this);
	}
	else
	{
		LOG_ERROR(_T("%s m_pcListener is NULL [%d, %d]"), __TFUNCTION__, GetIdx(), GetAccKey());
	}
}

/**
@brief 소켓이 재사용 되기 위한 초기화 작업을 한다.
*/
void CSession::Initialize()
{
	if(m_pSendIocpContext.get())
		m_pSendIocpContext->Init();
	else
		m_pSendIocpContext.reset(new CIocpContext(this, new CIocpEventSend)); 

	if(m_pRecvIocpContext.get())
		m_pRecvIocpContext->Init();
	else
		m_pRecvIocpContext.reset(new CIocpContext(this, new CIocpEventRecv(new CBuffer(MAX_CLIENT_PACKET_BUFFER_SIZE))));

	if(m_pAcceptIocpContext.get())
		m_pAcceptIocpContext->Init();
	else
		m_pAcceptIocpContext.reset(new CIocpContext(this, new CIocpEventAccept)); //버퍼를 잠시만 사용하지만 해당 포인터는 있어야되서 128짜리 만든거임.

	if(m_pCloseIocpContext.get())
		m_pCloseIocpContext->Init();
	else
		m_pCloseIocpContext.reset(new CIocpContext(this, new CIocpEventClose));

	m_bConnected = false;
	SetWorker(NULL);
}

/**
@brief 누적 키 값을 바꾼다. (재사용시 불려지는 메소드)
*/
void CSession::SetAccKey(const int &nAccKey)
{
	m_nAccKey = nAccKey;
}

/**
@brief 소켓을 닫는다. 
@desc 소켓을 닫으면 자동으로 IOCP에 걸려있던 이벤트가 return 된다.
*/
void CSession::CloseSocket()
{
	Enter();
	if(GetSocketRef() != INVALID_SOCKET)
		CSocketInterface::CloseSocket();
	Leave();
}

/**
@brief 기준틱과 비교를 통해 현재 틱이 유효한지 검사한다.
@param dwCurTick 현재 틱
@return 유효한지 여부
*/
bool CSession::CheckTick(DWORD dwCurTick)
{
	return m_pcTick->CheckTick(dwCurTick);
}

/**
@brief 기준 틱을 초기화해서 유효 시간을 늘린다.
*/
void CSession::ResetTick()
{
	m_pcTick->ResetTick();
}

/**
@brief Session와 연결된 개체들과의 정보를 초기화한다.
*/
void CSession::Release()
{
	LOG_DEBUG(_T("%s [%d] [%d, %d] [%p ,%p, %p, %p]"), __TFUNCTION__, GetSocket(), GetIdx(), GetAccKey(),
		GetRecvIocpContext(), GetSendIocpContext(), GetAcceptIocpContext(), GetCloseIocpContext());
	
	CScopeCriticalSection cs(&m_CriticalSection);
	m_nAccKey = -1;
	OnClose();
}

/**
@brief WSARecv를 걸어둠으로써, 데이터가 도착했을때 GQCS로 부터 리턴을 받을 수 있게 한다.
@return 성공 여부. 실패하면, 접속 종료 동작을 한다.
*/
bool CSession::Receive()
{
	CIocpContext* pIO = GetRecvIocpContext();
	if(pIO == NULL) return false;

	if(!isConnected()) 
		return false;
	
	DWORD dwRecv = 0, dwFlags = 0;
	int nError = WSARecv(GetSocket(), pIO->GetWsaBufPtr(), pIO->GetUsingWsaBufCount(), &dwRecv, &dwFlags,  pIO->GetOverlappedPtr(), NULL);
	if(nError == 0)
	{
		//한번에 받기 성공
		LOG_DEBUG(_T("%s Success 1 [%d, %d] [%d] [%p]"), __TFUNCTION__, GetIdx(), GetAccKey(), dwRecv, pIO->GetOverlappedPtr());
		return true;
	}
	else if(nError == SOCKET_ERROR && WSAGetLastError() == WSA_IO_PENDING)
	{
		//이것도 언젠간 성공이야
		LOG_DEBUG(_T("%s Success 2 [%d, %d] [%d] [%p]"), __TFUNCTION__, GetIdx(), GetAccKey(), dwRecv, pIO->GetOverlappedPtr());
		return true;
	}
	else
	{
		//실패야 ㅠㅠ
		LOG_INFO(_T("%s Fail [%d, %d, %d] [%d]"), __TFUNCTION__, GetIdx(), GetAccKey(), dwRecv, WSAGetLastError());
		m_pcWorker->OnIoFailed(this);
		return false;
	}
}

/**
@brief WSASend를 걸어둠으로써, 비동기로 데이터를 보낸다. 전송 완료했을때 GQCS로 부터 리턴을 받을 수 있게 한다.
@return 성공 여부. 실패하면, 접속 종료 동작을 한다.
@desc Lock으로 감싸고 들어와야함.
*/
bool CSession::Send()
{
	CIocpContext* pIO = GetSendIocpContext();
	if(pIO == NULL) return false;

	if(!isConnected()) 
		return false;

	if(!pIO->GetUsingWsaBufCount()) //쌓인 데이터 없으면 false
		return false;

	DWORD dwSend = 0, dwFlags = 0;
	int nError = WSASend(GetSocket(), pIO->GetWsaBufPtr(), pIO->GetUsingWsaBufCount(), &dwSend, 0, pIO->GetOverlappedPtr(), NULL);
	if(nError == 0)
	{
		//한번에 보내기 성공
		LOG_DEBUG(_T("%s Success [%d, %d] [%d] [%d]"), __TFUNCTION__, GetIdx(), GetAccKey(), dwSend);
		return true;
	} 
	else if(nError == SOCKET_ERROR && WSAGetLastError() == WSA_IO_PENDING)
	{
		//이것도 언젠간 성공이야
		LOG_DEBUG(_T("%s Success. WSA_IO_PENDING [%d, %d] [%d] [%d]"), __TFUNCTION__, GetIdx(), GetAccKey(), dwSend);
		return true;
	}
	else
	{
		//실패야 ㅠㅠ
		LOG_INFO(_T("%s Fail [%d] [%d, %d] [%d] [%p, %d]"), __TFUNCTION__, WSAGetLastError(), GetIdx(), GetAccKey(), dwSend, pIO->GetWsaBufPtr(),
			pIO->GetUsingWsaBufCount());
		m_pcWorker->OnIoFailed(this);
		return false;
	}
}

/**
@brief 데이터 전송 완료시 불려지는 메소드. Send 끝난 후 처리와, 필요시 다시 Send를 건다.
*/
void CSession::OnSend()
{
	CScopeCriticalSection cs(&m_CriticalSection);
	CIocpContext* pIO = GetSendIocpContext();
	if(pIO)
	{
		pIO->End();
		Send();
	}
}

/**
@brief 접속 완료시 불려지는 메소드.
*/
void CSession::OnConnect()
{
	ResetTick(); //접속되면 틱 초기화~
	InterlockedExchange((LONG *)(&m_bConnected), (LONG)true);
	if(m_pcPeer)
		m_pcPeer->OnConnect();

	LOG_DEBUG(_T("%s [%d, %d]"), __TFUNCTION__, GetIdx(), GetAccKey());
}

/**
@brief 접속 종료시 불려지는 메소드.
*/
void CSession::OnClose()
{
	InterlockedExchange((LONG *)(&m_bConnected), (LONG)false);
	if(m_pcPeer)
		m_pcPeer->OnClose();

	m_cProcessLock.Enter();
	Enter();
	m_pcPeer = NULL;
	m_pcWorker = NULL;
	Leave();
	m_cProcessLock.Leave();
	
	LOG_DEBUG(_T("%s [%d, %d]"), __TFUNCTION__, GetIdx(), GetAccKey());
}

/**
@brief 하나씩 쪼개진 패킷을 푸는 메소드.
*/
void CSession::PacketProcess(CBuffer* pcBuffer)
{
	m_cProcessLock.Enter();
	__try
	{
		if(m_pcPeer)
		{
			m_pcPeer->OnReceive(pcBuffer->GetUsingSize(), pcBuffer->GetHeadBuffer());
		}
	}
	__except(CMiniDumper::Handler(GetExceptionInformation()))
	{
		
	}
	m_cProcessLock.Leave();
}

/**
@brief Receive 완료시 불려지는 메소드. Peer에게 부탁해 (PacketParse를 통해) 패킷을 한개 단위로 쪼개서 큐에 넣어둔다.
@param nLen Receive된 데이터 길이
@param pData 데이터 포인터
@return 성공 여부
*/
bool CSession::OnReceive(unsigned int nLen, void* pData)
{
	CIocpContext* pIO = GetRecvIocpContext();
	CScopeCriticalSection cs(&m_CriticalSection);
	if(!isConnected())
		return false;

	pIO->Write(nLen);
	while(0 < pIO->GetUsingSize())
	{
		nLen = m_pcPeer->PacketParse(pIO->GetUsingSize(), pIO->GetHeadBuffer());
		if(0 == nLen)
			break;

		new CPacketQueue(GetIdx(), GetAccKey(), new CBuffer(nLen, pIO->GetHeadBuffer()));
		pIO->Read(nLen);
	}
	pIO->End();
	
	ResetTick(); //무조건 불러준다. 어떤 패킷이 오던 주기적으로 오면 살아있는 것이기에.
	return Receive(); //다음 통지 신호 받고 싶다고 등록
}

/**
@brief Send 요청시 불려지는 메소드. 
@param nLen Receive된 데이터 길이
@param pData 데이터 포인터
@param nAddLen pAddData의 크기
@param pAddData 기록할 추가 데이터
@return 성공 여부
*/
bool CSession::Send(const unsigned int nLen, const void* data, const unsigned int nAddLen, const void* pAddData)
{
	CIocpContext* pIO = GetSendIocpContext();
	CScopeCriticalSection cs(&m_CriticalSection);
	if(!isConnected())
		return false;

	ADDDATA_RESULT eResult = pIO->AddData(nLen, data, nAddLen, pAddData);
	if(m_pcPeer)
		m_pcPeer->Encrypt(pIO->GetUsingSize(), pIO->GetHeadBuffer());

	switch(eResult) 
	{
	case ADDDATARESULT_PENDING: //이후에 처리할 꺼란 얘기
		return true; //성공한거임!
	case ADDDATARESULT_FULL: //버퍼 꽉찼음
	case ADDDATARESULT_FAIL: //실패란 얘기
		Disconnect();
		return false;
	case ADDDATARESULT_SUCCESS:
		{
			bool bRet = Send();
			return bRet;
		}
	default:
		LOG_ERROR(_T("%s Error [%d]"), __TFUNCTION__, eResult);
		return false;
	}
}

/**
@brief Send 요청시 불려지는 메소드. 
@param nLen Receive된 데이터 길이
@param pData 데이터 포인터
@param nAddLen pAddData의 크기
@param pAddData 기록할 추가 데이터
@return 성공 여부
*/
void CSession::SetPeer(CPeer* pcPeer)
{
	m_pcPeer = pcPeer;
	m_pcPeer->SetIOInterface(this);
	m_pcPeer->SetSockAddrIn(&m_SockAddrIn);
}