﻿#include "stdafx.h"
#include "SessionManager.h"
#include "Session.h"
#include "Peer.h"

/**
@brief 소켓 컨텍스트 관리 클래스 생성자. 변수의 초기화를 담당한다.
*/
CSessionManager::CSessionManager() : m_nAccKey(0), m_nTotalCountSession(0)
{
}

/**
@brief 소켓 컨텍스트 관리 클래스 소멸자. 할당한 메모리를 해제한다.
*/
CSessionManager::~CSessionManager()
{
	while(!m_stl_map_UsingSession.empty()) 
	{
		STL_MAP_SESSION::iterator it = m_stl_map_UsingSession.begin();
		if(it != m_stl_map_UsingSession.end())
		{
			CSession* pcSession = it->second;
			Push(pcSession);
		}
	}

	while(!m_qNotUseSession.empty())
	{
		CSession* pcSession = m_qNotUseSession.front();
		m_qNotUseSession.pop();
		SAFE_DELETE(pcSession);
	}
}

/**
@brief 소켓 컨텍스트를 새로 할당한다. 
@param nCount 이 매개변수 값 만큼 새로 할당한다.
*/
void CSessionManager::New(int nCount)
{
	CScopeCriticalSection cs(&m_CriticalSection);
	for(int i = m_nTotalCountSession; i < nCount + m_nTotalCountSession; i++)
	{
		m_qNotUseSession.push(new CSession(i));
	}

	m_nTotalCountSession += nCount;
}

/**
@brief 소켓 컨텍스트를 하나 꺼내서 반환한다.
@return 사용 가능한 소켓 컨텍스트
@desc 없으면 하나 더 만들어서 리턴한다.
*/
CSession*  CSessionManager::Pop()
{
	CScopeCriticalSection cs(&m_CriticalSection);
	CSession* pcSession = NULL;
	if(m_qNotUseSession.empty()) 
	{
		LOG_INFO(_T("CSessionManager::Pop() m_qNotUseSession.empty()"));
		New(1); //하나 더 만들고 걔 쓰라고 함
		return Pop();
	}
	else
	{
		pcSession = m_qNotUseSession.front();
		m_qNotUseSession.pop();
	}

	pcSession->Initialize(); //기존에 사용되던 버퍼를 비워주고, 각종 데이터를 초기화해준다.
	pcSession->SetAccKey(m_nAccKey++); //누적키를 설정해준다.
	m_stl_map_UsingSession.insert(STL_MAP_SESSION::value_type(pcSession->GetIdx(), pcSession));
	LOG_DEBUG(_T("%s Ok [%d, %d]"), __TFUNCTION__, m_stl_map_UsingSession.size(), m_qNotUseSession.size());
	return pcSession;
}

/**
@brief 사용한 소켓 컨텍스트를 다시 큐에 집어넣는다
@param pcSession 큐에 접어넣을 소켓 컨텍스트
*/
void CSessionManager::Push(CSession* pcSession)
{
	CScopeCriticalSection cs(&m_CriticalSection);
	pcSession->Enter();
	pcSession->Release();
	m_stl_map_UsingSession.erase(pcSession->GetIdx()); //Get 해서 있었던거니 키로 지워도 된다고 확신!
	m_qNotUseSession.push(pcSession); //무조건 큐에 넣어 재활용하자.
	LOG_DEBUG(_T("%s [%d] [%d]"), __TFUNCTION__, m_stl_map_UsingSession.size(), m_qNotUseSession.size());
	pcSession->Leave();
}

/**
@brief 소켓 컨텍스트를 하나 찾아서 반환한다
@param nIdx 식별 값
@param nAccKey 식별용 누적 키 값
@return 성공 여부
*/
CSession* CSessionManager::Get(int nIdx, int nAccKey)
{
	CScopeCriticalSection cs(&m_CriticalSection);
	STL_MAP_SESSION::iterator it = m_stl_map_UsingSession.find(nIdx);
	if(it != m_stl_map_UsingSession.end())
	{
		CSession* pcSession = it->second;
		if(pcSession->GetAccKey() != nAccKey)
		{
			LOG_ERROR(_T("%s Acckey Different [%d] [%d, %d]"), __TFUNCTION__, nIdx, pcSession->GetAccKey(), nAccKey);
			return NULL;
		}
		return pcSession;
	}
	LOG_ERROR(_T("%s Not Find Session [%d, %d]"), __TFUNCTION__, nIdx, nAccKey);
	return NULL;
}

/**
@brief IO를 담당할 CPeer의 포인터를 소켓 컨텍스트에 등록한다.
@param nIdx 식별 값
@param nAccKey 식별용 누적 키 값
@param pcPeer 등록할 Peer
*/
void CSessionManager::SetPeer(int nIdx, int nAccKey, CPeer* pcPeer)
{
	CSession* pcSession = Get(nIdx, nAccKey);
	if(pcSession)
	{
		pcSession->SetPeer(pcPeer);
	}
}