﻿#pragma once

#include "Buffer.h"
#include <queue>
#include "CriticalSection.h"

///기본 WSASend 대기 큐 크기
const int DEFAULT_SEND_BUF_SIZE = 10;

///기본 WSASend에서 사용할 STL QUEUE 크기
const int DEFAULT_WAITING_STL_QUEUE_SIZE = 100;

class CIocpContext;
class CSession;

enum ADDDATA_RESULT
{
	ADDDATARESULT_NONE = -1,
	ADDDATARESULT_SUCCESS = 0, //성공
	ADDDATARESULT_PENDING = 1, //언젠간 처리 될것
	ADDDATARESULT_FAIL = 2, //실패
	ADDDATARESULT_FULL = 3, //꽉 찼음
};

/**
@brief IOCP에서 이벤트처리를 하는 데에 사용되는 기능을 담은 베이스 클래스.
*/
class CIocpEvent 
{
protected:
	///어떤 종류의 IO인지를 알기 위함
	IO_TYPE m_eIOType;
	
	///버퍼 포인터
	CBuffer* m_pcBuffer;

	///IOCP에서 사용되는 버퍼 형식 포인터
	WSABUF* m_pWsaBuf;

	///버퍼 크기
	int m_nWsaBufSize;

	//사장중인 버퍼 갯수
	int m_nUsingWsaBufCount;

protected:
	/**
	@brief 생성자.
	@param eIOType 어떤 종류의 IO를 담당할 객체인지
	@param pcBuffer IO에 사용될 버퍼 포인터
	@param nWsaBufSize 사용될 WSABUF 크기
	@desc 상속으로만 사용가능하게 하기 위해서 protected로 설정
	*/
	CIocpEvent(IO_TYPE eIOType, CBuffer* pcBuffer, int nWsaBufSize = 1) : m_nWsaBufSize(nWsaBufSize), 
		m_pWsaBuf(new WSABUF[nWsaBufSize]), m_eIOType(eIOType), m_pcBuffer(pcBuffer), m_nUsingWsaBufCount(0)
	{
		Init();
	}

public:
	/**
	@brief 소멸자.
	*/
	virtual ~CIocpEvent()
	{
		Release();
	}

	/**
	@brief 초기화 메소드
	*/
	virtual void Init()
	{
		for(int i = 0; i < m_nWsaBufSize; i++)
		{
			m_pWsaBuf[i].buf = NULL;
			m_pWsaBuf[i].len = 0;
		}
	}

	/**
	@brief 해제 메소드
	*/
	virtual void Release()
	{
		SAFE_DELETE(m_pcBuffer);
		SAFE_DELETE_ARRAY(m_pWsaBuf);
	}
	
	/**
	@brief 사용가능한 가장 앞 위치의 포인터를 반환하는 메소드
	@return 사용가능한 가장 앞 위치 포인터
	*/
	inline char* GetHeadBuffer()
	{
		if(m_pcBuffer)
		{
			return m_pcBuffer->GetHeadBuffer();
		}
		return NULL;
	}

	/**
	@brief 어떤 종류의 IO를 담당할지를 반환하는 메소드
	@return 어떤 종류의 IO를 담당하는지
	*/
	inline IO_TYPE GetIOType(){return m_eIOType;}

	/**
	@brief 버퍼의 크기를 알아오는 메소드
	@return 버퍼의 크기
	*/
	inline unsigned int GetSize(){return m_pcBuffer->GetSize();}

	/**
	@brief 버퍼에서 사용중인 크기를 알아오는 메소드
	@return 버퍼에서 사용중인 크기
	*/
	inline unsigned int GetUsingSize(){return m_pcBuffer->GetUsingSize();}

	/**
	@brief IO에 사용하는 WSABUF의 포인터를 반환하는 메소드
	@return IO에 사용하는 WSABUF의 포인터
	*/
	inline LPWSABUF GetWsaBufPtr(){return m_pWsaBuf;}

	/**
	@brief 사용중인 WSABUF의 갯수를 알아오는 메소드
	@return 사용중인 WSABUF의 갯수
	*/
	inline int GetUsingWsaBufCount(){return m_nUsingWsaBufCount;}

	/**
	@brief IO가 끝났을때 불려지는 메소드
	*/
	virtual void End(){}

	/**
	@brief 데이터 추가 메소드
	@param nLen pData의 크기
	@param pData 기록할 데이터
	@param nAddLen pAddData의 크기
	@param pAddData 기록할 추가 데이터
	@return 데이터 추가 결과. (재정의 하지 않으면 무조건 실패)
	@desc pData가 버퍼의 앞쪽, pAddData가 pData의 뒤쪽에 기록된다.
	*/
	virtual ADDDATA_RESULT AddData(const int len, const void* data, const int nAddLen = 0, const void* pAddData = 0) {return ADDDATARESULT_FAIL;}

	/**
	@brief 데이터를 얼만큼 읽었는지 전달하는 메소드.
	@param nLen 읽은 크기
	@return 성공 여부. (재정의 하지 않으면 무조건 실패)
	*/
	virtual bool Read(int nLen){return false;}

	/**
	@brief 데이터를 기록한 후 부르는 메소드
	@param nLen 기록한 크기
	@return 성공 여부. (재정의 하지 않으면 무조건 실패)
	*/
	virtual bool Write(int nLen){return false;}
};

/**
@brief IOCP의 Send에 이용되는 전담 클래스
*/
class CIocpEventSend : public CIocpEvent
{
	///기존 WSABUF가 사용중일때 쌓는 WSABUF 큐
	WSABUF* m_pWsaBufQueue;

	///큐 크기
	int m_nWsaBufQueueSize;

	///사용중인 큐 갯수
	int m_nQueueCount;

	///m_pWsaBufQueue가 꽉찼을 때 사용하는 큐
	std::queue<WSABUF> m_stl_queue_WsaBuf;

	///m_stl_queue_WsaBuf의 제한 갯수
	size_t m_nWaitingStlQueueSize;

public:
	/**
	@brief 생성자. 큐 크기를 전달 받는다.
	@param 큐 크기
	*/
	CIocpEventSend(int nWsaBufSize = DEFAULT_SEND_BUF_SIZE, size_t m_nWaitingStlQueueSize = DEFAULT_WAITING_STL_QUEUE_SIZE) : 
	  CIocpEvent(IOTYPE_SEND, NULL, nWsaBufSize), m_pWsaBufQueue(new WSABUF[nWsaBufSize]), m_nQueueCount(0),
		m_nWsaBufQueueSize(nWsaBufSize), m_nWaitingStlQueueSize(m_nWaitingStlQueueSize)
	{
		Init();
	}

	/**
	@brief 소멸자. Release 메소드를 호출한다.
	*/
	virtual ~CIocpEventSend()
	{
		Release();
	}

	/**
	@brief 초기화 메소드
	*/
	virtual void Init()
	{
		__super::Init();
		m_nUsingWsaBufCount = 0;

		for(int i = 0; i < m_nWsaBufQueueSize; i++)
		{
			m_pWsaBufQueue[i].buf = NULL;
			m_pWsaBufQueue[i].len = 0;
		}
		m_nQueueCount = 0;

		while(!m_stl_queue_WsaBuf.empty())
			m_stl_queue_WsaBuf.pop();
	}

	/**
	@brief 해제 메소드
	*/
	virtual void Release()
	{
		if(m_pcBuffer)
		{
			m_pcBuffer->Giveup();
			SAFE_DELETE(m_pcBuffer);
		}

		for(int i = 0 ; i < m_nUsingWsaBufCount; i++)
		{
			SAFE_DELETE_ARRAY(m_pWsaBuf[i].buf);
		}

		for(int i = 0 ; i < m_nWsaBufQueueSize; i++)
		{
			SAFE_DELETE_ARRAY(m_pWsaBufQueue[i].buf);
		}

		SAFE_DELETE_ARRAY(m_pWsaBufQueue);
		while(!m_stl_queue_WsaBuf.empty())
			m_stl_queue_WsaBuf.pop();

		__super::Release();
	}

	/**
	@brief IO 끝나고 스왑하는 메소드
	*/
	virtual void End()
	{
		//사용중이던 버퍼 반환
		for(int i = 0 ; i < m_nUsingWsaBufCount; i++)
		{
			SAFE_DELETE_ARRAY(m_pWsaBuf[i].buf);
		}

		m_nUsingWsaBufCount = 0; //버퍼 사용 카운트도 0으로 만든다.

		Swap<LPWSABUF>(m_pWsaBufQueue, m_pWsaBuf);
		Swap<int>(m_nUsingWsaBufCount, m_nQueueCount);
		Swap<int>(m_nWsaBufSize, m_nWsaBufQueueSize);

		while(!m_stl_queue_WsaBuf.empty())
		{
			WSABUF& refWsaBuf = m_stl_queue_WsaBuf.front();
			if(ADDDATARESULT_FAIL == Push(refWsaBuf))
				break;

			m_stl_queue_WsaBuf.pop();
		}
	}
	
	/**
	@brief 데이터 버퍼에 추가하는 메소드
	@param nLen pData의 크기
	@param pData 기록할 데이터
	@param nAddLen pAddData의 크기
	@param pAddData 기록할 추가 데이터
	@return 데이터 추가 결과
	@desc pData가 버퍼의 앞쪽, pAddData가 pData의 뒤쪽에 기록된다.
	*/
	virtual ADDDATA_RESULT AddData(const int nLen, const void* data, const int nAddLen = 0, const void* pAddData = 0)
	{
		if(m_pcBuffer)
		{
			m_pcBuffer->Giveup();
			SAFE_DELETE(m_pcBuffer);
		}

		m_pcBuffer = new CBuffer(nLen, data, nAddLen, pAddData);
		WSABUF wsaBuf;
		wsaBuf.len = m_pcBuffer->GetUsingSize();
		wsaBuf.buf = m_pcBuffer->GetHeadBuffer();

		if(GetUsingWsaBufCount() == 0)
		{
			m_pWsaBuf[m_nUsingWsaBufCount++] = wsaBuf;
			return ADDDATARESULT_SUCCESS;
		}

		if(ADDDATARESULT_FAIL == Push(wsaBuf))
		{
			if(m_stl_queue_WsaBuf.size() < m_nWaitingStlQueueSize)
				m_stl_queue_WsaBuf.push(wsaBuf);
			else
				return ADDDATARESULT_FULL;
		}

		return ADDDATARESULT_PENDING;
	}

	/**
	@brief 큐에 데이터를 쌓는 메소드
	@param refWsaBuf 큐에 쌓을 데이터
	@return 데이터 추가 결과
	*/
	ADDDATA_RESULT Push(WSABUF& refWsaBUF)
	{
		if(IS_VALID_VALUE(m_nQueueCount, 0, m_nWsaBufQueueSize))
		{
			m_pWsaBufQueue[m_nQueueCount++] = refWsaBUF;
			return ADDDATARESULT_PENDING;
		}
		return ADDDATARESULT_FAIL;
	}
};

/**
@brief IOCP의 Recv에 이용되는 전담 클래스
*/
class CIocpEventRecv : public CIocpEvent
{
public:
	/**
	@brief 생성자. 버퍼를 전달 받는다.
	@param pcBuffer 리시브에 사용할 버퍼
	*/
	CIocpEventRecv(CBuffer* pcBuffer) : CIocpEvent(IOTYPE_RECV, pcBuffer)
	{
		Init();
	}

	/**
	@brief 소멸자. 아무것도 안함.
	*/
	virtual ~CIocpEventRecv()
	{
	}

	/**
	@brief 초기화 메소드. 버퍼 초기화와 WSABUF를 세팅한다.
	*/
	virtual void Init()
	{
		m_pcBuffer->Init();
		BufferToWSABuf();
	}

	/**
	@brief 버퍼를 WSABUF에 설정한다.
	*/
	void BufferToWSABuf()
	{
		m_pWsaBuf[0].buf = m_pcBuffer->GetTailBuffer();
		m_pWsaBuf[0].len = m_pcBuffer->GetRemainSize();
		m_nUsingWsaBufCount = 1; //Recv는 무조건 1개 버퍼 사용하기 때문.
	}

	/**
	@brief IO완료시 처리. 버퍼를 압축하고, WSABUF로 세팅한다.
	*/
	virtual void End()
	{
		m_pcBuffer->Compress();
		BufferToWSABuf();
	}

	/**
	@brief 데이터 읽기 메소드
	@param nLen 읽은 크기
	@return 성공 여부
	*/
	virtual bool Read(int nLen)
	{
		return m_pcBuffer->JumpHead(nLen);
	}

	/**
	@brief 데이터 쓰기 메소드
	@param nLen 사용한 크기
	@return 성공 여부
	*/	
	virtual bool Write(int nLen)
	{
		return m_pcBuffer->JumpTail(nLen);
	}
};

/**
@brief IOCP의 Accept에 이용되는 전담 클래스
*/
class CIocpEventAccept : public CIocpEvent
{
public:
	/**
	@brief 생성자. 억셉트에 사용할 최소의 버퍼를 만들어 사용한다.
	*/
	CIocpEventAccept() : CIocpEvent(IOTYPE_ACCEPT, new CBuffer(ACCEPT_BUFFER_SIZE))
	{

	}

	/**
	@brief 소멸자. 아무것도 안함.
	*/
	virtual ~CIocpEventAccept()
	{
	}
};

/**
@brief IOCP의 Close에 이용되는 전담 클래스
*/
class CIocpEventClose : public CIocpEvent
{
public:
	/**
	@brief 생성자. 
	*/
	CIocpEventClose() : CIocpEvent(IOTYPE_CLOSE, NULL)
	{

	}

	/**
	@brief 소멸자. 아무것도 안함.
	*/
	virtual ~CIocpEventClose()
	{
	}

};

/**
@brief IOCP에 등록되고, 리턴되는 클래스.
*/
class CIocpContext //이 클래스는 가상 함수 만들면 안된다. 왜냐고? m_Overlapped의 위치가 가상 함수 테이블 크기만큼 밀려서 관리하기 어려워진다.
{
private:
	///비동기 IO를 위해 사용되는 Overlapped 객체
	OVERLAPPED m_Overlapped;

	///IO 전담 클래스 포인터
	CIocpEvent* m_pcIocpEvent;

	///나를 사용하는 CSession 포인터
	CSession* m_pcSession;

public:
	/**
	@brief 생성자.
	@param pcSession 나를 사용하는 CSession 포인터
	@param pcIocpEvent IO 전담 클래스 포인터
	*/
	CIocpContext(CSession* pcSession, CIocpEvent* pcIocpEvent) : m_pcSession(pcSession), m_pcIocpEvent(pcIocpEvent)
	{
		ZeroMemory(&m_Overlapped, sizeof(m_Overlapped));
	}

	/**
	@brief 소멸자. 동적할당한 데이터를 해제한다.
	*/
	~CIocpContext()
	{
		SAFE_DELETE(m_pcIocpEvent);
	}

	/**
	@brief 초기화 메소드
	*/
	inline void Init(){m_pcIocpEvent->Init();}

	/**
	@brief 비동기 IO를 위해 사용되는 Overlapped 객체 포인터를 반환하는 메소드
	@return 비동기 IO를 위해 사용되는 Overlapped 객체 포인터
	*/
	inline LPOVERLAPPED GetOverlappedPtr(){return &m_Overlapped;}

	/**
	@brief 소유하고 있는 IO 전담 클래스 포인터를 반환하는 메소드
	@return IO 전담 클래스 포인터
	*/
	inline CIocpEvent* GetIocpEvent(){return m_pcIocpEvent;}

	/**
	@brief 나를 사용하는 CSession 포인터를 반환하는 메소드
	@return 나를 사용하는 CSession 포인터
	*/
	inline CSession* GetSession(){return m_pcSession;}

	/**
	@brief 버퍼의 크기를 알아오는 메소드
	@return 버퍼의 크기
	*/
	inline unsigned int GetSize(){return m_pcIocpEvent->GetSize();}

	/**
	@brief 버퍼에서 사용중인 크기를 알아오는 메소드
	@return 버퍼에서 사용중인 크기
	*/
	inline unsigned int GetUsingSize(){return m_pcIocpEvent->GetUsingSize();}

	/**
	@brief 데이터를 얼만큼 읽었는지 전달하는 메소드.
	@param nLen 읽은 크기
	@return 성공 여부. 
	*/
	inline bool Read(int nLen){return m_pcIocpEvent->Read(nLen);}

	/**
	@brief 데이터를 기록한 후 부르는 메소드
	@param nLen 기록한 크기
	@return 성공 여부.
	*/
	inline bool Write(int nLen){return m_pcIocpEvent->Write(nLen);}

	/**
	@brief IO에 사용하는 WSABUF의 포인터를 반환하는 메소드
	@return IO에 사용하는 WSABUF의 포인터
	*/
	inline LPWSABUF GetWsaBufPtr(){return m_pcIocpEvent->GetWsaBufPtr();}

	/**
	@brief 데이터 추가 메소드
	@param nLen pData의 크기
	@param pData 기록할 데이터
	@param nAddLen pAddData의 크기
	@param pAddData 기록할 추가 데이터
	@return 데이터 추가 결과. (재정의 하지 않으면 무조건 실패)
	@desc pData가 버퍼의 앞쪽, pAddData가 pData의 뒤쪽에 기록된다.
	*/
	inline ADDDATA_RESULT AddData(const int len, const void* data, const int nAddLen = 0, const void* pAddData = 0) {return m_pcIocpEvent->AddData(len, data, nAddLen, pAddData);}

	/**
	@brief IO가 끝났을때 불려지는 메소드
	*/
	inline void End(){return m_pcIocpEvent->End();}

	/**
	@brief 사용가능한 가장 앞 위치의 포인터를 반환하는 메소드
	@return 사용가능한 가장 앞 위치 포인터
	*/
	inline char* GetHeadBuffer(){return m_pcIocpEvent->GetHeadBuffer();}

	/**
	@brief 어떤 종류의 IO를 담당할지를 반환하는 메소드
	@return 어떤 종류의 IO를 담당하는지
	*/
	inline IO_TYPE GetIOType(){return m_pcIocpEvent->GetIOType();}
	
	/**
	@brief 사용중인 WSABUF의 갯수를 알아오는 메소드
	@return 사용중인 WSABUF의 갯수
	*/
	inline int GetUsingWsaBufCount(){return m_pcIocpEvent->GetUsingWsaBufCount();}
};