﻿#pragma once

#include "Define.h"
#include "LogWriter.h"
#include <map>
#include "Singleton.h"
#include "Console.h"

#ifdef _UNICODE   
#define WIDEN(x)			L ## x   

#define WIDEN2(x)			WIDEN(x)   
#define __TFILE__			WIDEN2(__FILE__)   
#define __TFUNCTION__		WIDEN2(__FUNCTION__)   
#else   
#define __TFILE__			__FILE__   
#define __TFUNCTION__		__FUNCTION__   
#endif   

#define	LOGLEVEL_NONE	-1
#define	LOGLEVEL_DEBUG	0
#define	LOGLEVEL_INFO	1
#define	LOGLEVEL_ERROR	2
#define	LOGLEVEL_ALARM	3

enum LOG_MODE
{
	LOGMODE_NONE = 0,
	LOGMODE_CONSOLE,
	LOGMODE_DEBUGSTRING,
	MAX_LOGMODE,
};

/**
@brief 현재 실행중인 파일명을 얻어오는 메소드
@return 현재 실행중인 파일명
*/
inline tstring GetMoudleName()
{
	tstring strFileName;
	TCHAR szBuffer[MAX_PATH], *szFirstIdx;
	GetModuleFileName(NULL, szBuffer, MAX_PATH);
	szFirstIdx = _tcsrchr(szBuffer, _T('\\'));
	strFileName = szFirstIdx + 1;
	return strFileName;
}

/**
@brief 현재 실행중인 파일에서 확장자를 빼고 얻어오는 메소드
@return 현재 실행중인 파일명. (확장자 제외)
*/
inline tstring GetMoudleNameExclusiveExt()
{
	tstring strFileName = GetMoudleName();
	tstring::size_type nSizeType = strFileName.find_first_of(_T('.'));
	strFileName.at(nSizeType) = _T('\0');
	return strFileName;
}

/**
@brief Log의 활성화 상태와 이름을 담고 있는 구조체
*/
struct SLogState
{
	///로그가 켜져있는지 여부
	bool m_bOn;

	///로그의 이름
	tstring m_strName;

	/**
	@brief 생성자. 로그의 상태와 이름을 입력받아 저장한다.
	@param strName 로그의 이름
	@param bOn 켜져 있는지 여부
	*/
	SLogState(const tstring& strName, bool bOn = true) : m_bOn(bOn), m_strName(strName)
	{
	}
};

inline tstring GetTimeStampString(const TCHAR* szStr)
{
	SYSTEMTIME st;
	GetLocalTime(&st);
	tstringstream strStringStream;
	strStringStream << st.wHour << _T(":") << st.wMinute << _T(":") << st.wSecond<< " " << szStr << std::endl;
	return strStringStream.str();
}

/**
@brief Console창으로 로그를 출력하는 메소드
@param szStr 출력할 로그 문자열
*/
inline void ConsoleOutput(const TCHAR* szStr)
{
	tstring string = GetTimeStampString(szStr);
	HANDLE hOut = GetStdHandle(STD_OUTPUT_HANDLE);
	if( INVALID_HANDLE_VALUE != hOut )
	{
		DWORD dwWritten;
		::WriteConsole(hOut, string.c_str(), (DWORD)string.size(), &dwWritten, NULL);
	}
	else
	{
		::OutputDebugString(string.c_str());
	}
}

/**
@brief Console창으로 로그를 출력하는 메소드
@param szStr 출력할 로그 문자열
*/
inline void DebugStringOutput(const TCHAR* szStr)
{
	tstring string = GetTimeStampString(szStr);
	::OutputDebugString(string.c_str());
}

/**
@brief 로그 기능을 관리하는 관리자.
*/
class CLogManager
{
public:
	/**
	@brief 소멸자. 할당한 것들 해제.
	*/
	~CLogManager()
	{
	
	}

	/**
	@brief 로그 기록 모드를 콘솔 모드로 설정한다.
	*/
	inline void ConsoleMode(bool bOn = true)
	{
		if(bOn)
		{
			::AllocConsole();
			m_stl_map_LogCallBack.insert(STL_MAP_LOG_CALL_BACK::value_type(LOGMODE_CONSOLE, &ConsoleOutput));
		}
		else
		{
			::FreeConsole();
			m_stl_map_LogCallBack.erase(LOGMODE_CONSOLE);
		}
	}

	/**
	@brief 로그 기록 모드를 아웃풋 디버그 스트링 모드로 설정한다.
	*/
	inline void DebugStringMode(bool bOn = true)
	{
		if(bOn)
		{
			m_stl_map_LogCallBack.insert(STL_MAP_LOG_CALL_BACK::value_type(LOGMODE_DEBUGSTRING, &DebugStringOutput));
		}
		else
		{
			m_stl_map_LogCallBack.erase(LOGMODE_DEBUGSTRING);
		}
	}

	/**
	@brief 로그 상태 추가 메소드.
	@param dwLogFlag 로그 플래그
	@param refLogState 로그 참조자
	*/
	inline void Add(DWORD dwLogFlag, const SLogState& refLogState)
	{
		STL_MAP_LOG_STATE::iterator it = m_stl_map_LogFlag.find(dwLogFlag);
		if(it != m_stl_map_LogFlag.end())
			it->second = refLogState;
		else
			m_stl_map_LogFlag.insert(STL_MAP_LOG_STATE::value_type(dwLogFlag, refLogState));
	}

	/**
	@brief 로그 상태 변경 메소드.
	@param dwLogFlag 로그 플래그
	@param bOn OnOff 상태
	*/
	inline bool Set(DWORD dwLogFlag, bool bOn)
	{		
		SLogState* pcLogState = _Get(dwLogFlag);
		if(pcLogState)
		{
			pcLogState->m_bOn = bOn;
			return true;
		}
		return false;
	}

	/**
	@brief 등록된 디버그 함수 호출
	@param szStr 전달할 문자열
	*/
	inline void CallDebugFunc(const TCHAR* szStr)
	{
		for(STL_MAP_LOG_CALL_BACK::iterator it = m_stl_map_LogCallBack.begin(); it != m_stl_map_LogCallBack.end(); ++it)
		{
			(it->second)(szStr);
		}
	}

	/**
	@brief 로그 이름 설정
	@param strLogName 설정할 로그 이름
	*/
	inline void SetLogName(tstring strLogName)
	{
		m_strLogName = strLogName;
	}

	/**
	@brief 로그 기록 메소드.
	@param dwLogFlag 로그 플래그
	@param szStr 로그 문자열
	*/
	inline void Log(DWORD dwLogFlag, const TCHAR* szStr)
	{
		if(_isOn(dwLogFlag))
			CallDebugFunc(szStr);

		for(DWORD i = dwLogFlag; i != LOGLEVEL_NONE; i--)
			_Write(i, szStr);
	}

	///싱글턴으로 생성
	STATIC_SINGLETON_PTR(CLogManager);

private:
	/**
	@brief 로그 상태 얻어오는 메소드
	@param dwLogFlag 로그 플래그
	@return 로그 상태 구조체 변수 포인터. 등록된 상태 없으면 NULL.
	*/
	inline SLogState* _Get(DWORD dwLogFlag)
	{
		STL_MAP_LOG_STATE::iterator it = m_stl_map_LogFlag.find(dwLogFlag);
		if(it == m_stl_map_LogFlag.end())
			return NULL;

		return &it->second;
	}

	/**
	@brief 로그가 켜져있는지 여부
	@param dwLogFlag 로그 플래그
	@return 켜져있는지 여부. 등록되어있지 않다면 false.
	*/
	inline bool _isOn(DWORD dwLogFlag)
	{
		SLogState* pcLogState = _Get(dwLogFlag);
		if(pcLogState)
			return pcLogState->m_bOn;

		return false;
	}

	/**
	@brief 로그 기록 메소드
	@param dwLogFlag 로그 플래그
	@param szStr 로그 문자열
	*/
	inline void _Write(DWORD dwLogFlag, const TCHAR* szStr)
	{
		SLogState* pcLogState = _Get(dwLogFlag);
		if( NULL == pcLogState || !pcLogState->m_bOn )
			return;

		CLogWriter::GetInstancePtr()->Write(m_strLogName.c_str(), pcLogState->m_strName.c_str(), szStr);
	}

	/**
	@brief 생성자. 로그를 사용할때의 초기 기능을 설정한다.
	*/
	CLogManager()
	{
		SetLogName(GetMoudleNameExclusiveExt());
		ConsoleMode(true);
		CLogManager::GetInstancePtr()->Add(LOGLEVEL_DEBUG, SLogState(_T("DEBUG"), false));
		CLogManager::GetInstancePtr()->Add(LOGLEVEL_INFO, SLogState(_T("INFO")));
		CLogManager::GetInstancePtr()->Add(LOGLEVEL_ERROR, SLogState(_T("ERROR")));
		CLogManager::GetInstancePtr()->Add(LOGLEVEL_ALARM, SLogState(_T("ALARM")));
	}

private:
	///key : integer형 LogFlag. value : SLogState 구조체 변수
	typedef std::map<DWORD, SLogState> STL_MAP_LOG_STATE;

	///로그 상태를 관리하는 맵
	STL_MAP_LOG_STATE m_stl_map_LogFlag;

	///로그 이름
	tstring m_strLogName;

	///로그 콜백 함수의 typedef형
	typedef void (*LOG_CALL_BACK)(const TCHAR*) ; 

	///key: 로그 모드. value : LOG_CALL_BACK 함수 포인터
	typedef std::map<LOG_MODE, LOG_CALL_BACK> STL_MAP_LOG_CALL_BACK; 

	///로그 콜백을 담고 있는 맵
	STL_MAP_LOG_CALL_BACK m_stl_map_LogCallBack;
};

/**
@brief Debug 레벨로 로그 출력할 때 사용하는 클래스
*/
class CLogDebug : public CVariantArgumentString
{
protected:

	/**
	@brief Debug Level로 로그 출력하는 메소드. 회색 글씨로 출력.
	*/
	virtual void Write()
	{
		std::cout << con::fg_gray;
		CLogManager::GetInstancePtr()->Log(LOGLEVEL_DEBUG, GetString());
	}
};

/**
@brief Info 레벨로 로그 출력할 때 사용하는 클래스
*/
class CLogInfo : public CVariantArgumentString
{
protected:

	/**
	@brief Info Level로 로그 출력하는 메소드. 흰색 글씨로 출력.
	*/
	virtual void Write()
	{
		std::cout << con::fg_white;
		CLogManager::GetInstancePtr()->Log(LOGLEVEL_INFO, GetString());
	}
};

/**
@brief Error 레벨로 로그 출력할 때 사용하는 클래스
*/
class CLogError : public CVariantArgumentString
{
protected:

	/**
	@brief Error Level로 로그 출력하는 메소드. 빨간 글씨로 출력.
	*/
	virtual void Write()
	{
		std::cout << con::fg_red;
		CLogManager::GetInstancePtr()->Log(LOGLEVEL_ERROR, GetString());
	}	
};

/**
@brief Error 레벨로 로그 출력할 때 사용하는 클래스
*/
class CLogAlarm : public CVariantArgumentString
{
protected:

	/**
	@brief Error Level로 로그 출력하는 메소드. 빨간 글씨로 출력.
	*/
	virtual void Write()
	{
		std::cout << con::fg_cyan;
		CLogManager::GetInstancePtr()->Log(LOGLEVEL_ALARM, GetString());
	}	
};

///CLogDebug 형식으로 () 오퍼레이터 호출 하기 위한 매크로
#define LOG_DEBUG CLogDebug()

///CLogInfo 형식으로 () 오퍼레이터 호출 하기 위한 매크로
#define LOG_INFO CLogInfo()

///CLogError 형식으로 () 오퍼레이터 호출 하기 위한 매크로
#define LOG_ERROR CLogError()

///CLogAlarm 형식으로 () 오퍼레이터 호출 하기 위한 매크로
#define LOG_ALARM CLogAlarm()